@@ -3,7 +3,7 @@
 # Authors: Tom Kralidis <tomkralidis@gmail.com>
 #          Norman Barker <norman.barker@gmail.com>
 #
-# Copyright (c) 2022 Tom Kralidis
+# Copyright (c) 2024 Tom Kralidis
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation
@@ -31,23 +31,33 @@
 """Flask module providing the route paths to the api"""
 
 import os
+from typing import Union
 
 import click
+from flask import (Flask, Blueprint, make_response, request,
+                   send_from_directory, Response, Request)
 
-from flask import Flask, Blueprint, make_response, request, send_from_directory
-
-from pygeoapi.api import API
-from pygeoapi.util import get_mimetype, yaml_load, get_api_rules
+from pygeoapi.api import API, APIRequest, apply_gzip
+import pygeoapi.api.coverages as coverages_api
+import pygeoapi.api.environmental_data_retrieval as edr_api
+import pygeoapi.api.itemtypes as itemtypes_api
+import pygeoapi.api.maps as maps_api
+import pygeoapi.api.processes as processes_api
+import pygeoapi.api.stac as stac_api
+import pygeoapi.api.tiles as tiles_api
+from pygeoapi.openapi import load_openapi_document
+from pygeoapi.config import get_config
+from pygeoapi.util import get_mimetype, get_api_rules
 
 
-if 'PYGEOAPI_CONFIG' not in os.environ:
-    raise RuntimeError('PYGEOAPI_CONFIG environment variable not set')
-
-with open(os.environ.get('PYGEOAPI_CONFIG'), encoding='utf8') as fh:
-    CONFIG = yaml_load(fh)
+CONFIG = get_config()
+OPENAPI = load_openapi_document()
 
 API_RULES = get_api_rules(CONFIG)
 
+if CONFIG['server'].get('admin'):
+    from pygeoapi.admin import Admin
+
 STATIC_FOLDER = 'static'
 if 'templates' in CONFIG['server']:
     STATIC_FOLDER = CONFIG['server']['templates'].get('static', 'static')
@@ -61,6 +71,7 @@ BLUEPRINT = Blueprint(
     static_folder=STATIC_FOLDER,
     url_prefix=API_RULES.get_url_prefix('flask')
 )
+ADMIN_BLUEPRINT = Blueprint('admin', __name__, static_folder=STATIC_FOLDER)
 
 # CORS: optionally enable from config.
 if CONFIG['server'].get('cors', False):
@@ -73,7 +84,7 @@ if CONFIG['server'].get('cors', False):
 APP.config['JSONIFY_PRETTYPRINT_REGULAR'] = CONFIG['server'].get(
     'pretty_print', True)
 
-api_ = API(CONFIG)
+api_ = API(CONFIG, OPENAPI)
 
 OGC_SCHEMAS_LOCATION = CONFIG['server'].get('ogc_schemas_location')
 
@@ -98,12 +109,16 @@ if (OGC_SCHEMAS_LOCATION is not None and
         dirname_ = os.path.dirname(full_filepath)
         basename_ = os.path.basename(full_filepath)
 
-        # TODO: better sanitization?
-        path_ = dirname_.replace('..', '').replace('//', '')
+        path_ = dirname_.replace('..', '').replace('//', '').replace('./', '')
+
+        if '..' in path_:
+            return 'Invalid path', 400
+
         return send_from_directory(path_, basename_,
                                    mimetype=get_mimetype(basename_))
 
 
+# TODO: inline in execute_from_flask when all views have been refactored
 def get_response(result: tuple):
     """
     Creates a Flask Response object and updates matching headers.
@@ -111,7 +126,7 @@ def get_response(result: tuple):
     :param result: The result of the API call.
                    This should be a tuple of (headers, status, content).
 
-    :returns: A Response instance.
+    :returns: A Response instance
     """
 
     headers, status, content = result
@@ -122,6 +137,33 @@ def get_response(result: tuple):
     return response
 
 
+def execute_from_flask(api_function, request: Request, *args,
+                       skip_valid_check=False) -> Response:
+    """
+    Executes API function from Flask
+
+    :param api_function: API function
+    :param request: request object
+    :param *args: variable length additional arguments
+    :param skip_validity_check: bool
+
+    :returns: A Response instance
+    """
+
+    api_request = APIRequest.from_flask(request, api_.locales)
+
+    content: Union[str, bytes]
+
+    if not skip_valid_check and not api_request.is_valid():
+        headers, status, content = api_.get_format_exception(api_request)
+    else:
+        headers, status, content = api_function(api_, api_request, *args)
+        content = apply_gzip(headers, content)
+        # handle jsonld too?
+
+    return get_response((headers, status, content))
+
+
 @BLUEPRINT.route('/')
 def landing_page():
     """
@@ -139,13 +181,8 @@ def openapi():
 
     :returns: HTTP response
     """
-    with open(os.environ.get('PYGEOAPI_OPENAPI'), encoding='utf8') as ff:
-        if os.environ.get('PYGEOAPI_OPENAPI').endswith(('.yaml', '.yml')):
-            openapi_ = yaml_load(ff)
-        else:  # JSON string, do not transform
-            openapi_ = ff.read()
 
-    return get_response(api_.openapi(request, openapi_))
+    return get_response(api_.openapi_(request))
 
 
 @BLUEPRINT.route('/conformance')
@@ -155,9 +192,35 @@ def conformance():
 
     :returns: HTTP response
     """
+
     return get_response(api_.conformance(request))
 
 
+@BLUEPRINT.route('/TileMatrixSets/<tileMatrixSetId>')
+def get_tilematrix_set(tileMatrixSetId=None):
+    """
+    OGC API TileMatrixSet endpoint
+
+    :param tileMatrixSetId: identifier of tile matrix set
+
+    :returns: HTTP response
+    """
+
+    return execute_from_flask(tiles_api.tilematrixset, request,
+                              tileMatrixSetId)
+
+
+@BLUEPRINT.route('/TileMatrixSets')
+def get_tilematrix_sets():
+    """
+    OGC API TileMatrixSets endpoint
+
+    :returns: HTTP response
+    """
+
+    return execute_from_flask(tiles_api.tilematrixsets, request)
+
+
 @BLUEPRINT.route('/collections')
 @BLUEPRINT.route('/collections/<path:collection_id>')
 def collections(collection_id=None):
@@ -168,19 +231,35 @@ def collections(collection_id=None):
 
     :returns: HTTP response
     """
+
     return get_response(api_.describe_collections(request, collection_id))
 
 
-@BLUEPRINT.route('/collections/<path:collection_id>/queryables')
-def collection_queryables(collection_id=None):
+@BLUEPRINT.route('/collections/<path:collection_id>/schema')
+def collection_schema(collection_id):
     """
-    OGC API collections querybles endpoint
+    OGC API - collections schema endpoint
 
     :param collection_id: collection identifier
 
     :returns: HTTP response
     """
-    return get_response(api_.get_collection_queryables(request, collection_id))
+
+    return get_response(api_.get_collection_schema(request, collection_id))
+
+
+@BLUEPRINT.route('/collections/<path:collection_id>/queryables')
+def collection_queryables(collection_id=None):
+    """
+    OGC API collections queryables endpoint
+
+    :param collection_id: collection identifier
+
+    :returns: HTTP response
+    """
+
+    return execute_from_flask(itemtypes_api.get_collection_queryables, request,
+                              collection_id)
 
 
 @BLUEPRINT.route('/collections/<path:collection_id>/items',
@@ -201,36 +280,40 @@ def collection_items(collection_id, item_id=None):
 
     if item_id is None:
         if request.method == 'GET':  # list items
-            return get_response(
-                api_.get_collection_items(request, collection_id))
+            return execute_from_flask(itemtypes_api.get_collection_items,
+                                      request, collection_id,
+                                      skip_valid_check=True)
         elif request.method == 'POST':  # filter or manage items
             if request.content_type is not None:
                 if request.content_type == 'application/geo+json':
-                    return get_response(
-                        api_.manage_collection_item(request, 'create',
-                                                    collection_id))
+                    return execute_from_flask(
+                            itemtypes_api.manage_collection_item,
+                            request, 'create', collection_id,
+                            skip_valid_check=True)
                 else:
-                    return get_response(
-                        api_.post_collection_items(request, collection_id))
+                    return execute_from_flask(
+                            itemtypes_api.post_collection_items, request,
+                            collection_id, skip_valid_check=True)
         elif request.method == 'OPTIONS':
-            return get_response(
-                api_.manage_collection_item(request, 'options', collection_id))
+            return execute_from_flask(
+                    itemtypes_api.manage_collection_item, request, 'options',
+                    collection_id, skip_valid_check=True)
 
     elif request.method == 'DELETE':
-        return get_response(
-            api_.manage_collection_item(request, 'delete',
-                                        collection_id, item_id))
+        return execute_from_flask(itemtypes_api.manage_collection_item,
+                                  request, 'delete', collection_id, item_id,
+                                  skip_valid_check=True)
     elif request.method == 'PUT':
-        return get_response(
-            api_.manage_collection_item(request, 'update',
-                                        collection_id, item_id))
+        return execute_from_flask(itemtypes_api.manage_collection_item,
+                                  request, 'update', collection_id, item_id,
+                                  skip_valid_check=True)
     elif request.method == 'OPTIONS':
-        return get_response(
-            api_.manage_collection_item(request, 'options',
-                                        collection_id, item_id))
+        return execute_from_flask(itemtypes_api.manage_collection_item,
+                                  request, 'options', collection_id, item_id,
+                                  skip_valid_check=True)
     else:
-        return get_response(
-            api_.get_collection_item(request, collection_id, item_id))
+        return execute_from_flask(itemtypes_api.get_collection_item, request,
+                                  collection_id, item_id)
 
 
 @BLUEPRINT.route('/collections/<path:collection_id>/coverage')
@@ -242,33 +325,9 @@ def collection_coverage(collection_id):
 
     :returns: HTTP response
     """
-    return get_response(api_.get_collection_coverage(request, collection_id))
 
-
-@BLUEPRINT.route('/collections/<path:collection_id>/coverage/domainset')
-def collection_coverage_domainset(collection_id):
-    """
-    OGC API - Coverages coverage domainset endpoint
-
-    :param collection_id: collection identifier
-
-    :returns: HTTP response
-    """
-    return get_response(api_.get_collection_coverage_domainset(
-        request, collection_id))
-
-
-@BLUEPRINT.route('/collections/<path:collection_id>/coverage/rangetype')
-def collection_coverage_rangetype(collection_id):
-    """
-    OGC API - Coverages coverage rangetype endpoint
-
-    :param collection_id: collection identifier
-
-    :returns: HTTP response
-    """
-    return get_response(api_.get_collection_coverage_rangetype(
-        request, collection_id))
+    return execute_from_flask(coverages_api.get_collection_coverage, request,
+                              collection_id, skip_valid_check=True)
 
 
 @BLUEPRINT.route('/collections/<path:collection_id>/tiles')
@@ -280,8 +339,9 @@ def get_collection_tiles(collection_id=None):
 
     :returns: HTTP response
     """
-    return get_response(api_.get_collection_tiles(
-        request, collection_id))
+
+    return execute_from_flask(tiles_api.get_collection_tiles, request,
+                              collection_id)
 
 
 @BLUEPRINT.route('/collections/<path:collection_id>/tiles/<tileMatrixSetId>')
@@ -295,8 +355,10 @@ def get_collection_tiles_metadata(collection_id=None, tileMatrixSetId=None):
 
     :returns: HTTP response
     """
-    return get_response(api_.get_collection_tiles_metadata(
-        request, collection_id, tileMatrixSetId))
+
+    return execute_from_flask(tiles_api.get_collection_tiles_metadata,
+                              request, collection_id, tileMatrixSetId,
+                              skip_valid_check=True)
 
 
 @BLUEPRINT.route('/collections/<path:collection_id>/tiles/\
@@ -314,8 +376,12 @@ def get_collection_tiles_data(collection_id=None, tileMatrixSetId=None,
 
     :returns: HTTP response
     """
-    return get_response(api_.get_collection_tiles_data(
-        request, collection_id, tileMatrixSetId, tileMatrix, tileRow, tileCol))
+
+    return execute_from_flask(
+        tiles_api.get_collection_tiles_data,
+        request, collection_id, tileMatrixSetId, tileMatrix, tileRow, tileCol,
+        skip_valid_check=True,
+    )
 
 
 @BLUEPRINT.route('/collections/<collection_id>/map')
@@ -330,19 +396,13 @@ def collection_map(collection_id, style_id=None):
     :returns: HTTP response
     """
 
-    headers, status_code, content = api_.get_collection_map(
-        request, collection_id, style_id)
-
-    response = make_response(content, status_code)
-
-    if headers:
-        response.headers = headers
-
-    return response
+    return execute_from_flask(
+        maps_api.get_collection_map, request, collection_id, style_id
+    )
 
 
 @BLUEPRINT.route('/processes')
-@BLUEPRINT.route('/processes/<path:process_id>')
+@BLUEPRINT.route('/processes/<process_id>')
 def get_processes(process_id=None):
     """
     OGC API - Processes description endpoint
@@ -351,7 +411,9 @@ def get_processes(process_id=None):
 
     :returns: HTTP response
     """
-    return get_response(api_.describe_processes(request, process_id))
+
+    return execute_from_flask(processes_api.describe_processes, request,
+                              process_id)
 
 
 @BLUEPRINT.route('/jobs')
@@ -367,15 +429,16 @@ def get_jobs(job_id=None):
     """
 
     if job_id is None:
-        return get_response(api_.get_jobs(request))
+        return execute_from_flask(processes_api.get_jobs, request)
     else:
         if request.method == 'DELETE':  # dismiss job
-            return get_response(api_.delete_job(job_id))
+            return execute_from_flask(processes_api.delete_job, request,
+                                      job_id)
         else:  # Return status of a specific job
-            return get_response(api_.get_jobs(request, job_id))
+            return execute_from_flask(processes_api.get_jobs, request, job_id)
 
 
-@BLUEPRINT.route('/processes/<path:process_id>/execution', methods=['POST'])
+@BLUEPRINT.route('/processes/<process_id>/execution', methods=['POST'])
 def execute_process_jobs(process_id):
     """
     OGC API - Processes execution endpoint
@@ -385,7 +448,8 @@ def execute_process_jobs(process_id):
     :returns: HTTP response
     """
 
-    return get_response(api_.execute_process(request, process_id))
+    return execute_from_flask(processes_api.execute_process, request,
+                              process_id)
 
 
 @BLUEPRINT.route('/jobs/<job_id>/results',
@@ -398,7 +462,8 @@ def get_job_result(job_id=None):
 
     :returns: HTTP response
     """
-    return get_response(api_.get_job_result(request, job_id))
+
+    return execute_from_flask(processes_api.get_job_result, request, job_id)
 
 
 @BLUEPRINT.route('/jobs/<job_id>/results/<resource>',
@@ -412,6 +477,8 @@ def get_job_result_resource(job_id, resource):
 
     :returns: HTTP response
     """
+
+    # TODO: this does not seem to exist?
     return get_response(api_.get_job_result_resource(
         request, job_id, resource))
 
@@ -422,24 +489,38 @@ def get_job_result_resource(job_id, resource):
 @BLUEPRINT.route('/collections/<path:collection_id>/radius')
 @BLUEPRINT.route('/collections/<path:collection_id>/trajectory')
 @BLUEPRINT.route('/collections/<path:collection_id>/corridor')
+@BLUEPRINT.route('/collections/<path:collection_id>/locations/<location_id>')  # noqa
+@BLUEPRINT.route('/collections/<path:collection_id>/locations')  # noqa
 @BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/position')  # noqa
 @BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/area')  # noqa
 @BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/cube')  # noqa
 @BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/radius')  # noqa
 @BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/trajectory')  # noqa
 @BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/corridor')  # noqa
-def get_collection_edr_query(collection_id, instance_id=None):
+@BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/locations/<location_id>')  # noqa
+@BLUEPRINT.route('/collections/<path:collection_id>/instances/<instance_id>/locations')  # noqa
+def get_collection_edr_query(collection_id, instance_id=None,
+                             location_id=None):
     """
     OGC EDR API endpoints
 
     :param collection_id: collection identifier
     :param instance_id: instance identifier
+    :param location_id: location id of a /locations/<location_id> query
 
     :returns: HTTP response
     """
-    query_type = request.path.split('/')[-1]
-    return get_response(api_.get_collection_edr_query(request, collection_id,
-                                                      instance_id, query_type))
+
+    if location_id:
+        query_type = 'locations'
+    else:
+        query_type = request.path.split('/')[-1]
+
+    return execute_from_flask(
+        edr_api.get_collection_edr_query, request, collection_id, instance_id,
+        query_type, location_id,
+        skip_valid_check=True,
+    )
 
 
 @BLUEPRINT.route('/stac')
@@ -449,7 +530,8 @@ def stac_catalog_root():
 
     :returns: HTTP response
     """
-    return get_response(api_.get_stac_root(request))
+
+    return execute_from_flask(stac_api.get_stac_root, request)
 
 
 @BLUEPRINT.route('/stac/<path:path>')
@@ -461,11 +543,72 @@ def stac_catalog_path(path):
 
     :returns: HTTP response
     """
-    return get_response(api_.get_stac_path(request, path))
+
+    return execute_from_flask(stac_api.get_stac_path, request, path)
+
+
+@ADMIN_BLUEPRINT.route('/admin/config', methods=['GET', 'PUT', 'PATCH'])
+def admin_config():
+    """
+    Admin endpoint
+
+    :returns: HTTP response
+    """
+
+    if request.method == 'GET':
+        return get_response(admin_.get_config(request))
+
+    elif request.method == 'PUT':
+        return get_response(admin_.put_config(request))
+
+    elif request.method == 'PATCH':
+        return get_response(admin_.patch_config(request))
+
+
+@ADMIN_BLUEPRINT.route('/admin/config/resources', methods=['GET', 'POST'])
+def admin_config_resources():
+    """
+    Resources endpoint
+
+    :returns: HTTP response
+    """
+
+    if request.method == 'GET':
+        return get_response(admin_.get_resources(request))
+
+    elif request.method == 'POST':
+        return get_response(admin_.post_resource(request))
+
+
+@ADMIN_BLUEPRINT.route(
+    '/admin/config/resources/<resource_id>',
+    methods=['GET', 'PUT', 'PATCH', 'DELETE'])
+def admin_config_resource(resource_id):
+    """
+    Resource endpoint
+
+    :returns: HTTP response
+    """
+
+    if request.method == 'GET':
+        return get_response(admin_.get_resource(request, resource_id))
+
+    elif request.method == 'DELETE':
+        return get_response(admin_.delete_resource(request, resource_id))
+
+    elif request.method == 'PUT':
+        return get_response(admin_.put_resource(request, resource_id))
+
+    elif request.method == 'PATCH':
+        return get_response(admin_.patch_resource(request, resource_id))
 
 
 APP.register_blueprint(BLUEPRINT)
 
+if CONFIG['server'].get('admin'):
+    admin_ = Admin(CONFIG, OPENAPI)
+    APP.register_blueprint(ADMIN_BLUEPRINT)
+
 
 @click.command()
 @click.pass_context
