@@ -1,9 +1,10 @@
 # =================================================================
 #
 # Authors: Tom Kralidis <tomkralidis@gmail.com>
+#          Francesco Bartoli <xbartolone@gmail.com>
 #
 # Copyright (c) 2023 Tom Kralidis
-# Copyright (c) 2021 Francesco Bartoli
+# Copyright (c) 2024 Francesco Bartoli
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation
@@ -99,7 +100,7 @@ class ElasticsearchProvider(BaseProvider):
         """
 
         fields_ = {}
-        ii = self.es.indices.get(index=self.index_name)
+        ii = self.es.indices.get(index=self.index_name, allow_no_indices=False)
 
         LOGGER.debug(f'Response: {ii}')
         try:
@@ -294,7 +295,7 @@ class ElasticsearchProvider(BaseProvider):
         try:
             LOGGER.debug('querying Elasticsearch')
             if filterq:
-                LOGGER.debug(f'adding cql object: {filterq.json()}')
+                LOGGER.debug(f'adding cql object: {filterq.model_dump_json()}')
                 query = update_query(input_query=query, cql=filterq)
             LOGGER.debug(json.dumps(query, indent=4))
 
@@ -463,9 +464,10 @@ class ElasticsearchProvider(BaseProvider):
         LOGGER.debug('Fetching id and geometry from GeoJSON document')
         feature_ = doc['_source']
 
-        if self.id_field in doc['_source']['properties']:
+        try:
             id_ = doc['_source']['properties'][self.id_field]
-        else:
+        except KeyError as err:
+            LOGGER.debug(f'Missing field: {err}')
             id_ = doc['_source'].get('id', doc['_id'])
 
         feature_['id'] = id_
@@ -513,9 +515,9 @@ class ElasticsearchProvider(BaseProvider):
         if not self.properties and not self.select_properties:
             all_properties = self.get_fields()
         if self.properties and self.select_properties:
-            all_properties = set(self.properties) & set(self.select_properties)
+            all_properties = self.properties and self.select_properties
         else:
-            all_properties = set(self.properties) | set(self.select_properties)
+            all_properties = self.properties or self.select_properties
 
         LOGGER.debug(f'resulting properties: {all_properties}')
         return all_properties
@@ -645,16 +647,16 @@ class ESQueryBuilder:
 def _build_query(q, cql):
 
     # this would be handled by the AST with the traverse of CQL model
-    op, node = get_next_node(cql.__root__)
+    op, node = get_next_node(cql.root)
     q.operation = op
     if isinstance(node, list):
         query_list = []
         for elem in node:
             op, next_node = get_next_node(elem)
             if not getattr(next_node, 'between', 0) == 0:
-                property = next_node.between.value.__root__.__root__.property
-                lower = next_node.between.lower.__root__.__root__
-                upper = next_node.between.upper.__root__.__root__
+                property = next_node.between.value.root.root.property
+                lower = next_node.between.lower.root.root
+                upper = next_node.between.upper.root.root
                 query_list.append(Q(
                     {
                         'range':
@@ -665,24 +667,24 @@ def _build_query(q, cql):
                             }
                     }
                 ))
-            if not getattr(next_node, '__root__', 0) == 0:
-                scalars = tuple(next_node.__root__.eq.__root__)
-                property = scalars[0].__root__.property
-                value = scalars[1].__root__.__root__
+            if not getattr(next_node, 'root', 0) == 0:
+                scalars = tuple(next_node.root.eq.root)
+                property = scalars[0].root.property
+                value = scalars[1].root.root
                 query_list.append(Q(
                     {'match': {f'{property}': f'{value}'}}
                 ))
         q.must(query_list)
     elif not getattr(node, 'between', 0) == 0:
-        property = node.between.value.__root__.__root__.property
+        property = node.between.value.root.root.property
         lower = None
         if not getattr(node.between.lower,
-                       '__root__', 0) == 0:
-            lower = node.between.lower.__root__.__root__
+                       'root', 0) == 0:
+            lower = node.between.lower.root.root
         upper = None
         if not getattr(node.between.upper,
-                       '__root__', 0) == 0:
-            upper = node.between.upper.__root__.__root__
+                       'root', 0) == 0:
+            upper = node.between.upper.root.root
         query = Q(
             {
                 'range':
@@ -694,26 +696,26 @@ def _build_query(q, cql):
             }
         )
         q.must(query)
-    elif not getattr(node, '__root__', 0) == 0:
+    elif not getattr(node, 'root', 0) == 0:
         next_op, next_node = get_next_node(node)
         if not getattr(next_node, 'eq', 0) == 0:
-            scalars = tuple(next_node.eq.__root__)
-            property = scalars[0].__root__.property
-            value = scalars[1].__root__.__root__
+            scalars = tuple(next_node.eq.root)
+            property = scalars[0].root.property
+            value = scalars[1].root.root
             query = Q(
                 {'match': {f'{property}': f'{value}'}}
             )
             q.must(query)
     elif not getattr(node, 'intersects', 0) == 0:
-        property = node.intersects.__root__[0].__root__.property
+        property = node.intersects.root[0].root.property
         if property == 'geometry':
-            geom_type = node.intersects.__root__[
-                1].__root__.__root__.__root__.type
-            if geom_type.value == 'Polygon':
-                coordinates = node.intersects.__root__[
-                    1].__root__.__root__.__root__.coordinates
+            geom_type = node.intersects.root[
+                1].root.root.root.type
+            if geom_type == 'Polygon':
+                coordinates = node.intersects.root[
+                    1].root.root.root.coordinates
                 coords_list = [
-                    poly_coords.__root__ for poly_coords in coordinates[0]
+                    poly_coords.root for poly_coords in coordinates[0]
                 ]
                 filter_ = Q(
                     {
