@@ -5,7 +5,7 @@
 #          Abdulazeez Abdulazeez Adeshina <youngestdev@gmail.com>
 #
 # Copyright (c) 2020 Francesco Bartoli
-# Copyright (c) 2022 Tom Kralidis
+# Copyright (c) 2024 Tom Kralidis
 # Copyright (c) 2022 Abdulazeez Abdulazeez Adeshina
 #
 # Permission is hereby granted, free of charge, to any person
@@ -32,12 +32,12 @@
 # =================================================================
 """ Starlette module providing the route paths to the api"""
 
+import asyncio
 import os
-from typing import Union
+from typing import Callable, Union
 from pathlib import Path
 
 import click
-
 from starlette.routing import Route, Mount
 from starlette.staticfiles import StaticFiles
 from starlette.applications import Starlette
@@ -49,14 +49,27 @@ from starlette.responses import (
 )
 import uvicorn
 
-from pygeoapi.api import API
-from pygeoapi.util import yaml_load, get_api_rules
+from pygeoapi.api import API, APIRequest, apply_gzip
+import pygeoapi.api.coverages as coverages_api
+import pygeoapi.api.environmental_data_retrieval as edr_api
+import pygeoapi.api.itemtypes as itemtypes_api
+import pygeoapi.api.maps as maps_api
+import pygeoapi.api.processes as processes_api
+import pygeoapi.api.stac as stac_api
+import pygeoapi.api.tiles as tiles_api
+from pygeoapi.openapi import load_openapi_document
+from pygeoapi.config import get_config
+from pygeoapi.util import get_api_rules
 
-if 'PYGEOAPI_CONFIG' not in os.environ:
-    raise RuntimeError('PYGEOAPI_CONFIG environment variable not set')
+CONFIG = get_config()
 
-with open(os.environ.get('PYGEOAPI_CONFIG'), encoding='utf8') as fh:
-    CONFIG = yaml_load(fh)
+if 'PYGEOAPI_OPENAPI' not in os.environ:
+    raise RuntimeError('PYGEOAPI_OPENAPI environment variable not set')
+
+OPENAPI = load_openapi_document()
+
+if CONFIG['server'].get('admin'):
+    from pygeoapi.admin import Admin
 
 p = Path(__file__)
 
@@ -70,20 +83,48 @@ except KeyError:
 
 API_RULES = get_api_rules(CONFIG)
 
-api_ = API(CONFIG)
+api_ = API(CONFIG, OPENAPI)
 
 
-def get_response(result: tuple) -> Union[Response, JSONResponse, HTMLResponse]:
+def call_api_threadsafe(
+    loop: asyncio.AbstractEventLoop, api_call: Callable, *args
+) -> tuple:
+    """
+    The api call needs a running loop. This method is meant to be called
+    from a thread that has no loop running.
+
+    :param loop: The loop to use.
+    :param api_call: The API method to call.
+    :param args: Arguments to pass to the API method.
+    :returns: The api call result tuple.
+    """
+    asyncio.set_event_loop(loop)
+    return api_call(*args)
+
+
+async def get_response(
+        api_call,
+        *args,
+) -> Union[Response, JSONResponse, HTMLResponse]:
     """
     Creates a Starlette Response object and updates matching headers.
 
+    Runs the core api handler in a separate thread in order to avoid
+    blocking the main event loop.
+
     :param result: The result of the API call.
                    This should be a tuple of (headers, status, content).
 
     :returns: A Response instance.
     """
 
-    headers, status, content = result
+    loop = asyncio.get_running_loop()
+    headers, status, content = await loop.run_in_executor(
+        None, call_api_threadsafe, loop, api_call, *args)
+    return _to_response(headers, status, content)
+
+
+def _to_response(headers, status, content):
     if headers['Content-Type'] == 'text/html':
         response = HTMLResponse(content=content, status_code=status)
     else:
@@ -97,6 +138,27 @@ def get_response(result: tuple) -> Union[Response, JSONResponse, HTMLResponse]:
     return response
 
 
+async def execute_from_starlette(api_function, request: Request, *args,
+                                 skip_valid_check=False) -> Response:
+    api_request = await APIRequest.from_starlette(request, api_.locales)
+    content: Union[str, bytes]
+    if not skip_valid_check and not api_request.is_valid():
+        headers, status, content = api_.get_format_exception(api_request)
+    else:
+
+        loop = asyncio.get_running_loop()
+        headers, status, content = await loop.run_in_executor(
+            None, call_api_threadsafe, loop, api_function,
+            api_, api_request, *args)
+        # NOTE: that gzip currently doesn't work in starlette
+        #       https://github.com/geopython/pygeoapi/issues/1591
+        content = apply_gzip(headers, content)
+
+    response = _to_response(headers, status, content)
+
+    return response
+
+
 async def landing_page(request: Request):
     """
     OGC API landing page endpoint
@@ -105,7 +167,7 @@ async def landing_page(request: Request):
 
     :returns: Starlette HTTP Response
     """
-    return get_response(api_.landing_page(request))
+    return await get_response(api_.landing_page, request)
 
 
 async def openapi(request: Request):
@@ -116,13 +178,7 @@ async def openapi(request: Request):
 
     :returns: Starlette HTTP Response
     """
-    with open(os.environ.get('PYGEOAPI_OPENAPI'), encoding='utf8') as ff:
-        if os.environ.get('PYGEOAPI_OPENAPI').endswith(('.yaml', '.yml')):
-            openapi_ = yaml_load(ff)
-        else:  # JSON file, do not transform
-            openapi_ = ff
-
-    return get_response(api_.openapi(request, openapi_))
+    return await get_response(api_.openapi_, request)
 
 
 async def conformance(request: Request):
@@ -133,7 +189,47 @@ async def conformance(request: Request):
 
     :returns: Starlette HTTP Response
     """
-    return get_response(api_.conformance(request))
+    return await get_response(api_.conformance, request)
+
+
+async def get_tilematrix_set(request: Request, tileMatrixSetId=None):
+    """
+    OGC API TileMatrixSet endpoint
+
+    :param tileMatrixSetId: identifier of tile matrix set
+    :returns: HTTP response
+    """
+    if 'tileMatrixSetId' in request.path_params:
+        tileMatrixSetId = request.path_params['tileMatrixSetId']
+
+    return await execute_from_starlette(
+        tiles_api.tilematrixset, request, tileMatrixSetId,
+    )
+
+
+async def get_tilematrix_sets(request: Request):
+    """
+    OGC API TileMatrixSets endpoint
+
+    :returns: HTTP response
+    """
+    return await execute_from_starlette(tiles_api.tilematrixsets, request)
+
+
+async def collection_schema(request: Request, collection_id=None):
+    """
+    OGC API collections schema endpoint
+
+    :param request: Starlette Request instance
+    :param collection_id: collection identifier
+
+    :returns: Starlette HTTP Response
+    """
+    if 'collection_id' in request.path_params:
+        collection_id = request.path_params['collection_id']
+
+    return await get_response(api_.get_collection_schema, request,
+                              collection_id)
 
 
 async def collection_queryables(request: Request, collection_id=None):
@@ -147,7 +243,10 @@ async def collection_queryables(request: Request, collection_id=None):
     """
     if 'collection_id' in request.path_params:
         collection_id = request.path_params['collection_id']
-    return get_response(api_.get_collection_queryables(request, collection_id))
+
+    return await execute_from_starlette(
+        itemtypes_api.get_collection_queryables, request, collection_id,
+    )
 
 
 async def get_collection_tiles(request: Request, collection_id=None):
@@ -161,8 +260,9 @@ async def get_collection_tiles(request: Request, collection_id=None):
     """
     if 'collection_id' in request.path_params:
         collection_id = request.path_params['collection_id']
-    return get_response(api_.get_collection_tiles(
-        request, collection_id))
+
+    return await execute_from_starlette(
+        tiles_api.get_collection_tiles, request, collection_id)
 
 
 async def get_collection_tiles_metadata(request: Request, collection_id=None,
@@ -179,8 +279,11 @@ async def get_collection_tiles_metadata(request: Request, collection_id=None,
         collection_id = request.path_params['collection_id']
     if 'tileMatrixSetId' in request.path_params:
         tileMatrixSetId = request.path_params['tileMatrixSetId']
-    return get_response(api_.get_collection_tiles_metadata(
-        request, collection_id, tileMatrixSetId))
+
+    return await execute_from_starlette(
+        tiles_api.get_collection_tiles_metadata, request,
+        collection_id, tileMatrixSetId, skip_valid_check=True,
+    )
 
 
 async def get_collection_items_tiles(request: Request, collection_id=None,
@@ -208,9 +311,11 @@ async def get_collection_items_tiles(request: Request, collection_id=None,
         tileRow = request.path_params['tileRow']
     if 'tileCol' in request.path_params:
         tileCol = request.path_params['tileCol']
-    return get_response(api_.get_collection_tiles_data(
-        request, collection_id, tileMatrixSetId,
-        tile_matrix, tileRow, tileCol))
+    return await execute_from_starlette(
+        tiles_api.get_collection_tiles_data, request, collection_id,
+        tileMatrixSetId, tile_matrix, tileRow, tileCol,
+        skip_valid_check=True,
+    )
 
 
 async def collection_items(request: Request, collection_id=None, item_id=None):
@@ -230,38 +335,47 @@ async def collection_items(request: Request, collection_id=None, item_id=None):
         item_id = request.path_params['item_id']
     if item_id is None:
         if request.method == 'GET':  # list items
-            return get_response(
-                api_.get_collection_items(
-                    request, collection_id))
+            return await execute_from_starlette(
+                itemtypes_api.get_collection_items, request, collection_id,
+                skip_valid_check=True)
         elif request.method == 'POST':  # filter or manage items
             content_type = request.headers.get('content-type')
             if content_type is not None:
                 if content_type == 'application/geo+json':
-                    return get_response(
-                        api_.manage_collection_item(request, 'create',
-                                                    collection_id))
+                    return await execute_from_starlette(
+                        itemtypes_api.manage_collection_item, request,
+                        'create', collection_id, skip_valid_check=True)
                 else:
-                    return get_response(
-                        api_.post_collection_items(request, collection_id))
+                    return await execute_from_starlette(
+                        itemtypes_api.post_collection_items,
+                        request,
+                        collection_id,
+                        skip_valid_check=True,
+                    )
         elif request.method == 'OPTIONS':
-            return get_response(
-                api_.manage_collection_item(request, 'options', collection_id))
+            return await execute_from_starlette(
+                itemtypes_api.manage_collection_item, request,
+                'options', collection_id, skip_valid_check=True,
+            )
 
     elif request.method == 'DELETE':
-        return get_response(
-            api_.manage_collection_item(request, 'delete',
-                                        collection_id, item_id))
+        return await execute_from_starlette(
+            itemtypes_api.manage_collection_item, request, 'delete',
+            collection_id, item_id, skip_valid_check=True,
+        )
     elif request.method == 'PUT':
-        return get_response(
-            api_.manage_collection_item(request, 'update',
-                                        collection_id, item_id))
+        return await execute_from_starlette(
+            itemtypes_api.manage_collection_item, request, 'update',
+            collection_id, item_id, skip_valid_check=True,
+        )
     elif request.method == 'OPTIONS':
-        return get_response(
-            api_.manage_collection_item(request, 'options',
-                                        collection_id, item_id))
+        return await execute_from_starlette(
+            itemtypes_api.manage_collection_item, request, 'options',
+            collection_id, item_id, skip_valid_check=True,
+        )
     else:
-        return get_response(api_.get_collection_item(
-            request, collection_id, item_id))
+        return await execute_from_starlette(
+            itemtypes_api.get_collection_item, request, collection_id, item_id)
 
 
 async def collection_coverage(request: Request, collection_id=None):
@@ -276,40 +390,9 @@ async def collection_coverage(request: Request, collection_id=None):
     if 'collection_id' in request.path_params:
         collection_id = request.path_params['collection_id']
 
-    return get_response(api_.get_collection_coverage(request, collection_id))
-
-
-async def collection_coverage_domainset(request: Request, collection_id=None):
-    """
-    OGC API - Coverages coverage domainset endpoint
-
-    :param request: Starlette Request instance
-    :param collection_id: collection identifier
-
-    :returns: Starlette HTTP Response
-    """
-    if 'collection_id' in request.path_params:
-        collection_id = request.path_params['collection_id']
-
-    return get_response(api_.get_collection_coverage_domainset(
-        request, collection_id))
-
-
-async def collection_coverage_rangetype(request: Request, collection_id=None):
-    """
-    OGC API - Coverages coverage rangetype endpoint
-
-    :param request: Starlette Request instance
-    :param collection_id: collection identifier
-
-    :returns: Starlette HTTP Response
-    """
-
-    if 'collection_id' in request.path_params:
-        collection_id = request.path_params['collection_id']
-
-    return get_response(api_.get_collection_coverage_rangetype(
-        request, collection_id))
+    return await execute_from_starlette(
+        coverages_api.get_collection_coverage, request, collection_id,
+        skip_valid_check=True)
 
 
 async def collection_map(request: Request, collection_id, style_id=None):
@@ -327,8 +410,9 @@ async def collection_map(request: Request, collection_id, style_id=None):
     if 'style_id' in request.path_params:
         style_id = request.path_params['style_id']
 
-    return get_response(api_.get_collection_map(
-        request, collection_id, style_id))
+    return await execute_from_starlette(
+        maps_api.get_collection_map, request, collection_id, style_id
+    )
 
 
 async def get_processes(request: Request, process_id=None):
@@ -343,7 +427,8 @@ async def get_processes(request: Request, process_id=None):
     if 'process_id' in request.path_params:
         process_id = request.path_params['process_id']
 
-    return get_response(api_.describe_processes(request, process_id))
+    return await execute_from_starlette(processes_api.describe_processes,
+                                        request, process_id)
 
 
 async def get_jobs(request: Request, job_id=None):
@@ -360,12 +445,14 @@ async def get_jobs(request: Request, job_id=None):
         job_id = request.path_params['job_id']
 
     if job_id is None:  # list of submit job
-        return get_response(api_.get_jobs(request))
+        return await execute_from_starlette(processes_api.get_jobs, request)
     else:  # get or delete job
         if request.method == 'DELETE':
-            return get_response(api_.delete_job(job_id))
+            return await execute_from_starlette(processes_api.delete_job,
+                                                request, job_id)
         else:  # Return status of a specific job
-            return get_response(api_.get_jobs(request, job_id))
+            return await execute_from_starlette(processes_api.get_jobs,
+                                                request, job_id)
 
 
 async def execute_process_jobs(request: Request, process_id=None):
@@ -381,7 +468,8 @@ async def execute_process_jobs(request: Request, process_id=None):
     if 'process_id' in request.path_params:
         process_id = request.path_params['process_id']
 
-    return get_response(api_.execute_process(request, process_id))
+    return await execute_from_starlette(processes_api.execute_process,
+                                        request, process_id)
 
 
 async def get_job_result(request: Request, job_id=None):
@@ -397,7 +485,8 @@ async def get_job_result(request: Request, job_id=None):
     if 'job_id' in request.path_params:
         job_id = request.path_params['job_id']
 
-    return get_response(api_.get_job_result(request, job_id))
+    return await execute_from_starlette(processes_api.get_job_result,
+                                        request, job_id)
 
 
 async def get_job_result_resource(request: Request,
@@ -417,8 +506,9 @@ async def get_job_result_resource(request: Request,
     if 'resource' in request.path_params:
         resource = request.path_params['resource']
 
-    return get_response(api_.get_job_result_resource(
-        request, job_id, resource))
+    # TODO: this api function currently doesn't exist
+    return await get_response(
+        api_.get_job_result_resource, request, job_id, resource)
 
 
 async def get_collection_edr_query(request: Request, collection_id=None, instance_id=None):  # noqa
@@ -438,8 +528,11 @@ async def get_collection_edr_query(request: Request, collection_id=None, instanc
         instance_id = request.path_params['instance_id']
 
     query_type = request["path"].split('/')[-1]  # noqa
-    return get_response(api_.get_collection_edr_query(request, collection_id,
-                                                      instance_id, query_type))
+    return await execute_from_starlette(
+        edr_api.get_collection_edr_query, request, collection_id,
+        instance_id, query_type,
+        skip_valid_check=True,
+    )
 
 
 async def collections(request: Request, collection_id=None):
@@ -453,7 +546,8 @@ async def collections(request: Request, collection_id=None):
     """
     if 'collection_id' in request.path_params:
         collection_id = request.path_params['collection_id']
-    return get_response(api_.describe_collections(request, collection_id))
+    return await get_response(
+        api_.describe_collections, request, collection_id)
 
 
 async def stac_catalog_root(request: Request):
@@ -464,7 +558,7 @@ async def stac_catalog_root(request: Request):
 
     :returns: Starlette HTTP response
     """
-    return get_response(api_.get_stac_root(request))
+    return await execute_from_starlette(stac_api.get_stac_root, request)
 
 
 async def stac_catalog_path(request: Request):
@@ -476,7 +570,61 @@ async def stac_catalog_path(request: Request):
     :returns: Starlette HTTP response
     """
     path = request.path_params["path"]
-    return get_response(api_.get_stac_path(request, path))
+    return await execute_from_starlette(stac_api.get_stac_path, request, path)
+
+
+async def admin_config(request: Request):
+    """
+    Admin endpoint
+
+    :returns: Starlette HTTP Response
+    """
+
+    if request.method == 'GET':
+        return await get_response(ADMIN.get_config, request)
+    elif request.method == 'PUT':
+        return await get_response(ADMIN.put_config, request)
+    elif request.method == 'PATCH':
+        return await get_response(ADMIN.patch_config, request)
+
+
+async def admin_config_resources(request: Request):
+    """
+    Resources endpoint
+
+    :returns: HTTP response
+    """
+
+    if request.method == 'GET':
+        return await get_response(ADMIN.get_resources, request)
+    elif request.method == 'POST':
+        return await get_response(ADMIN.put_resource, request)
+
+
+async def admin_config_resource(request: Request, resource_id: str):
+    """
+    Resource endpoint
+
+    :param resource_id: resource identifier
+
+    :returns: Starlette HTTP Response
+    """
+
+    if 'resource_id' in request.path_params:
+        resource_id = request.path_params['resource_id']
+
+    if request.method == 'GET':
+        return await get_response(
+            ADMIN.get_resource, request, resource_id)
+    elif request.method == 'PUT':
+        return await get_response(
+            ADMIN.put_resource, request, resource_id)
+    elif request.method == 'PATCH':
+        return await get_response(
+            ADMIN.patch_resource, request, resource_id)
+    elif request.method == 'DELETE':
+        return await get_response(
+            ADMIN.delete_resource, request, resource_id)
 
 
 class ApiRulesMiddleware:
@@ -516,6 +664,9 @@ api_routes = [
     Route('/', landing_page),
     Route('/openapi', openapi),
     Route('/conformance', conformance),
+    Route('/TileMatrixSets/{tileMatrixSetId}', get_tilematrix_set),
+    Route('/TileMatrixSets', get_tilematrix_sets),
+    Route('/collections/{collection_id:path}/schema', collection_schema),
     Route('/collections/{collection_id:path}/queryables', collection_queryables),  # noqa
     Route('/collections/{collection_id:path}/tiles', get_collection_tiles),
     Route('/collections/{collection_id:path}/tiles/{tileMatrixSetId}', get_collection_tiles_metadata),  # noqa
@@ -524,8 +675,6 @@ api_routes = [
     Route('/collections/{collection_id:path}/items', collection_items, methods=['GET', 'POST', 'OPTIONS']),  # noqa
     Route('/collections/{collection_id:path}/items/{item_id:path}', collection_items, methods=['GET', 'PUT', 'DELETE', 'OPTIONS']),  # noqa
     Route('/collections/{collection_id:path}/coverage', collection_coverage),  # noqa
-    Route('/collections/{collection_id:path}/coverage/domainset', collection_coverage_domainset),  # noqa
-    Route('/collections/{collection_id:path}/coverage/rangetype', collection_coverage_rangetype),  # noqa
     Route('/collections/{collection_id:path}/map', collection_map),
     Route('/collections/{collection_id:path}/styles/{style_id:path}/map', collection_map),  # noqa
     Route('/processes', get_processes),
@@ -541,18 +690,33 @@ api_routes = [
     Route('/collections/{collection_id:path}/radius', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/trajectory', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/corridor', get_collection_edr_query),  # noqa
+    Route('/collections/{collection_id:path}/locations', get_collection_edr_query),  # noqa
+    Route('/collections/{collection_id:path}/locations/{location_id}', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/instances/{instance_id}/position', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/instances/{instance_id}/area', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/instances/{instance_id}/cube', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/instances/{instance_id}/radius', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/instances/{instance_id}/trajectory', get_collection_edr_query),  # noqa
     Route('/collections/{collection_id:path}/instances/{instance_id}/corridor', get_collection_edr_query),  # noqa
+    Route('/collections/{collection_id:path}/instances/{instance_id}/locations', get_collection_edr_query),  # noqa
+    Route('/collections/{collection_id:path}/instances/{instance_id}/locations/{location_id}', get_collection_edr_query),  # noqa
     Route('/collections', collections),
     Route('/collections/{collection_id:path}', collections),
     Route('/stac', stac_catalog_root),
     Route('/stac/{path:path}', stac_catalog_path),
 ]
 
+admin_routes = [
+    Route('/admin/config', admin_config, methods=['GET', 'PUT', 'PATCH']),
+    Route('/admin/config/resources', admin_config_resources, methods=['GET', 'POST']),  # noqa
+    Route('/admin/config/resources/{resource_id:path}', admin_config_resource,
+          methods=['GET', 'PUT', 'PATCH', 'DELETE'])
+]
+
+if CONFIG['server'].get('admin', False):
+    ADMIN = Admin(CONFIG, OPENAPI)
+    api_routes.extend(admin_routes)
+
 url_prefix = API_RULES.get_url_prefix('starlette')
 APP = Starlette(
     routes=[
@@ -613,7 +777,7 @@ def serve(ctx, server=None, debug=False):
     if debug:
         log_level = 'debug'
     uvicorn.run(
-        "pygeoapi.starlette_app:app",
+        "pygeoapi.starlette_app:APP",
         reload=True,
         log_level=log_level,
         loop='asyncio',
