@@ -30,25 +30,30 @@
 """Generic util functions used in the code"""
 
 import base64
-import json
-import logging
-import mimetypes
-import os
-import re
+from copy import deepcopy
+from filelock import FileLock
 import functools
 from functools import partial
 from dataclasses import dataclass
 from datetime import date, datetime, time
 from decimal import Decimal
 from enum import Enum
+import json
+import logging
+import mimetypes
+import os
+import pathlib
 from pathlib import Path
-from typing import Any, IO, Union, List, Callable
+import re
+from typing import Any, IO, Union, List, Optional, Callable
 from urllib.parse import urlparse
 from urllib.request import urlopen
+import uuid
 
 import dateutil.parser
-import shapely.ops
+from shapely import ops
 from shapely.geometry import (
+    box,
     GeometryCollection,
     LinearRing,
     LineString,
@@ -63,6 +68,8 @@ from shapely.geometry import (
 import yaml
 from babel.support import Translations
 from jinja2 import Environment, FileSystemLoader, select_autoescape
+import pygeofilter.ast
+import pygeofilter.values
 import pyproj
 from pyproj.exceptions import CRSError
 from requests import Session
@@ -78,22 +85,8 @@ LOGGER = logging.getLogger(__name__)
 
 DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
 
-TEMPLATES = Path(__file__).parent.resolve() / 'templates'
-
-CRS_AUTHORITY = [
-    "AUTO",
-    "EPSG",
-    "OGC",
-]
-
-# Global to compile only once
-CRS_URI_PATTERN = re.compile(
-    (
-     rf"^http://www.opengis\.net/def/crs/"
-     rf"(?P<auth>{'|'.join(CRS_AUTHORITY)})/"
-     rf"[\d|\.]+?/(?P<code>\w+?)$"
-    )
-)
+THISDIR = Path(__file__).parent.resolve()
+TEMPLATES = THISDIR / 'templates'
 
 
 # Type for Shapely geometrical objects.
@@ -171,23 +164,39 @@ def yaml_load(fh: IO) -> dict:
     :returns: `dict` representation of YAML
     """
 
-    # support environment variables in config
-    # https://stackoverflow.com/a/55301129
-    path_matcher = re.compile(r'.*\$\{([^}^{]+)\}.*')
+    # # support environment variables in config
+    # # https://stackoverflow.com/a/55301129
 
-    def path_constructor(loader, node):
-        env_var = path_matcher.match(node.value).group(1)
-        if env_var not in os.environ:
-            msg = f'Undefined environment variable {env_var} in config'
-            raise EnvironmentError(msg)
-        return get_typed_value(os.path.expandvars(node.value))
+    env_matcher = re.compile(
+        r'.*?\$\{(?P<varname>\w+)(:-(?P<default>[^}]+))?\}')
+
+    def env_constructor(loader, node):
+        result = ""
+        current_index = 0
+        raw_value = node.value
+        for match_obj in env_matcher.finditer(raw_value):
+            groups = match_obj.groupdict()
+            varname_start = match_obj.span('varname')[0]
+            result += raw_value[current_index:(varname_start-2)]
+            if (var_value := os.getenv(groups['varname'])) is not None:
+                result += var_value
+            elif (default_value := groups.get('default')) is not None:
+                result += default_value
+            else:
+                raise EnvironmentError(
+                    f'Could not find the {groups["varname"]!r} environment '
+                    f'variable'
+                )
+            current_index = match_obj.end()
+        else:
+            result += raw_value[current_index:]
+        return get_typed_value(result)
 
     class EnvVarLoader(yaml.SafeLoader):
         pass
 
-    EnvVarLoader.add_implicit_resolver('!path', path_matcher, None)
-    EnvVarLoader.add_constructor('!path', path_constructor)
-
+    EnvVarLoader.add_implicit_resolver('!env', env_matcher, None)
+    EnvVarLoader.add_constructor('!env', env_constructor)
     return yaml.load(fh, Loader=EnvVarLoader)
 
 
@@ -208,6 +217,32 @@ def get_base_url(config: dict) -> str:
     return url_join(config['server']['url'], rules.get_url_prefix())
 
 
+def yaml_dump(dict_: dict, destfile: str) -> bool:
+    """
+    Dump dict to YAML file
+
+    :param dict_: `dict` to dump
+    :param destfile: destination filepath
+
+    :returns: `bool`
+    """
+
+    def path_representer(dumper, data):
+        return dumper.represent_scalar(u'tag:yaml.org,2002:str', str(data))
+
+    yaml.add_multi_representer(pathlib.PurePath, path_representer)
+
+    lock = FileLock(f'{destfile}.lock')
+
+    with lock:
+        LOGGER.debug('Dumping YAML document')
+        with open(destfile, 'wb') as fh:
+            yaml.dump(dict_, fh, sort_keys=False, encoding='utf8', indent=4,
+                      default_flow_style=False)
+
+    return True
+
+
 def str2bool(value: Union[bool, str]) -> bool:
     """
     helper function to return Python boolean
@@ -358,16 +393,20 @@ def json_serial(obj: Any) -> str:
             return base64.b64encode(obj)
     elif isinstance(obj, Decimal):
         return float(obj)
-    elif type(obj).__name__ == 'int64':
+    elif type(obj).__name__ in ['int32', 'int64']:
         return int(obj)
-    elif type(obj).__name__ == 'float64':
+    elif type(obj).__name__ in ['float32', 'float64']:
         return float(obj)
     elif isinstance(obj, l10n.Locale):
         return l10n.locale2str(obj)
-
-    msg = f'{obj} type {type(obj)} not serializable'
-    LOGGER.error(msg)
-    raise TypeError(msg)
+    elif isinstance(obj, (pathlib.PurePath, Path)):
+        return str(obj)
+    elif isinstance(obj, uuid.UUID):
+        return str(obj)
+    else:
+        msg = f'{obj} type {type(obj)} not serializable'
+        LOGGER.error(msg)
+        raise TypeError(msg)
 
 
 def is_url(urlstring: str) -> bool:
@@ -571,6 +610,19 @@ class JobStatus(Enum):
     dismissed = 'dismissed'
 
 
+@dataclass(frozen=True)
+class Subscriber:
+    """
+    Store subscriber URLs as defined in:
+
+    https://schemas.opengis.net/ogcapi/processes/part1/1.0/openapi/schemas/subscriber.yaml  # noqa
+    """
+
+    success_uri: str
+    in_progress_uri: Optional[str]
+    failed_uri: Optional[str]
+
+
 def read_data(path: Union[Path, str]) -> Union[bytes, str]:
     """
     helper function to read data (file or network)
@@ -654,9 +706,9 @@ def get_crs_from_uri(uri: str) -> pyproj.CRS:
     Author: @MTachon
 
     :param uri: Uniform resource identifier of the coordinate
-        reference system.
-    :type uri: str
-
+                reference system. In accordance with
+                https://docs.ogc.org/pol/09-048r5.html#_naming_rule URIs can
+                take either the form of a URL or a URN
     :raises `CRSError`: Error raised if no CRS could be identified from the
         URI.
 
@@ -664,23 +716,26 @@ def get_crs_from_uri(uri: str) -> pyproj.CRS:
     :rtype: `pyproj.CRS`
     """
 
+    # normalize the input `uri` to a URL first
+    url = uri.replace(
+        "urn:ogc:def:crs",
+        "http://www.opengis.net/def/crs"
+    ).replace(":", "/")
     try:
-        crs = pyproj.CRS.from_authority(*CRS_URI_PATTERN.search(uri).groups())
-    except CRSError:
+        authority, code = url.rsplit("/", maxsplit=3)[1::2]
+        crs = pyproj.CRS.from_authority(authority, code)
+    except ValueError:
         msg = (
-            f"CRS could not be identified from URI {uri!r} "
-            f"(Authority: {CRS_URI_PATTERN.search(uri).group('auth')!r}, "
-            f"Code: {CRS_URI_PATTERN.search(uri).group('code')!r})."
+            f"CRS could not be identified from URI {uri!r}. CRS URIs must "
+            "follow one of two formats: "
+            "'http://www.opengis.net/def/crs/{authority}/{version}/{code}' or "
+            "'urn:ogc:def:crs:{authority}:{version}:{code}' "
+            "(see https://docs.opengeospatial.org/is/18-058r1/18-058r1.html#crs-overview)."  # noqa
         )
         LOGGER.error(msg)
         raise CRSError(msg)
-    except AttributeError:
-        msg = (
-            f"CRS could not be identified from URI {uri!r}. CRS URIs must "
-            "follow the format "
-            "'http://www.opengis.net/def/crs/{authority}/{version}/{code}' "
-            "(see https://docs.opengeospatial.org/is/18-058r1/18-058r1.html#crs-overview)."  # noqa
-        )
+    except CRSError:
+        msg = f"CRS could not be identified from URI {uri!r}"
         LOGGER.error(msg)
         raise CRSError(msg)
     else:
@@ -710,7 +765,7 @@ def get_transform_from_crs(
     crs_transform = pyproj.Transformer.from_crs(
         crs_in, crs_out, always_xy=always_xy,
     ).transform
-    return partial(shapely.ops.transform, crs_transform)
+    return partial(ops.transform, crs_transform)
 
 
 def crs_transform(func):
@@ -821,6 +876,7 @@ class UrlPrefetcher:
     """ Prefetcher to get HTTP headers for specific URLs.
     Allows a maximum of 1 redirect by default.
     """
+
     def __init__(self):
         self._session = Session()
         self._session.max_redirects = 1
@@ -840,3 +896,142 @@ class UrlPrefetcher:
         except Exception:  # noqa
             return CaseInsensitiveDict()
         return response.headers
+
+
+def bbox2geojsongeometry(bbox: list) -> dict:
+    """
+    Converts bbox values into GeoJSON geometry
+
+    :param bbox: `list` of minx, miny, maxx, maxy
+
+    :returns: `dict` of GeoJSON geometry
+    """
+
+    b = box(*bbox, ccw=False)
+    return geom_to_geojson(b)
+
+
+def modify_pygeofilter(
+        ast_tree: pygeofilter.ast.Node,
+        *,
+        filter_crs_uri: str,
+        storage_crs_uri: Optional[str] = None,
+        geometry_column_name: Optional[str] = None
+) -> pygeofilter.ast.Node:
+    """
+    Modifies the input pygeofilter with information from the provider.
+
+    :param ast_tree: `pygeofilter.ast.Node` representing the
+                     already parsed pygeofilter expression
+    :param filter_crs_uri: URI of the CRS being used in the filtering
+                           expression
+    :param storage_crs_uri: An optional string containing the URI of
+                            the provider's storage CRS
+    :param geometry_column_name: An optional string containing the
+                                 actual name of the provider's geometry field
+    :returns: A new pygeofilter.ast.Node, with the modified filter
+              expression
+
+    This function modifies the parsed pygeofilter that contains the raw
+    filter expression provided by an external client. It performs the
+    following modifications:
+
+    - if the filter includes any spatial coordinates and they are being
+      provided in a different CRS from the provider's storage CRS, the
+      corresponding geometries are transformed into the storage CRS
+
+    - if the filter includes the generic 'geometry' name as a reference to
+      the actual geometry of features, it is replaced by the actual name
+      of the geometry field, as specified by the provider
+
+    """
+    new_tree = deepcopy(ast_tree)
+    if storage_crs_uri:
+        storage_crs = get_crs_from_uri(storage_crs_uri)
+        filter_crs = get_crs_from_uri(filter_crs_uri)
+        _inplace_transform_filter_geometries(new_tree, filter_crs, storage_crs)
+    if geometry_column_name:
+        _inplace_replace_geometry_filter_name(new_tree, geometry_column_name)
+    return new_tree
+
+
+def _inplace_transform_filter_geometries(
+        node: pygeofilter.ast.Node,
+        filter_crs: pyproj.CRS,
+        storage_crs: pyproj.CRS
+):
+    """
+    Recursively traverse node tree and convert coordinates to the storage CRS.
+
+    This function modifies nodes in the already-parsed filter in order to find
+    any geometry literals that may be used in the filter and, if necessary,
+    proceeds to convert spatial coordinates to the CRS used by the provider.
+    """
+    try:
+        sub_nodes = node.get_sub_nodes()
+    except AttributeError:
+        pass
+    else:
+        for sub_node in sub_nodes:
+            is_geometry_node = isinstance(
+                sub_node, pygeofilter.values.Geometry)
+            if is_geometry_node:
+                # NOTE1: To be flexible, and since pygeofilter
+                # already supports it, in addition to supporting
+                # the `filter-crs` parameter, we also support having a
+                # geometry defined in EWKT, meaning the CRS is provided
+                # inline, like this `SRID=<CRS_CODE>;<WKT>` - If provided,
+                # this overrides the value of `filter-crs`. This enables
+                # supporting, for example, an exotic filter expression with
+                # multiple geometries specified in different CRSs
+
+                # NOTE2: We specify a default CRS using a URI of type URN
+                # because this is what pygeofilter uses internally too
+
+                crs_urn_provided_in_ewkt = sub_node.geometry.get(
+                    'crs', {}).get('properties', {}).get('name')
+                if crs_urn_provided_in_ewkt is not None:
+                    crs = get_crs_from_uri(crs_urn_provided_in_ewkt)
+                else:
+                    crs = filter_crs
+                if crs != storage_crs:
+                    # convert geometry coordinates to storage crs
+                    geom = geojson_to_geom(sub_node.geometry)
+                    coord_transformer = pyproj.Transformer.from_crs(
+                        crs_from=crs, crs_to=storage_crs).transform
+                    transformed_geom = ops.transform(coord_transformer, geom)
+                    sub_node.geometry = geom_to_geojson(transformed_geom)
+                # ensure the crs is encoded in the sub-node, otherwise
+                # pygeofilter will assign it its own default CRS
+                authority, code = storage_crs.to_authority()
+                sub_node.geometry['crs'] = {
+                    'properties': {
+                        'name': f'urn:ogc:def:crs:{authority}::{code}'
+                    }
+                }
+            else:
+                _inplace_transform_filter_geometries(
+                    sub_node, filter_crs, storage_crs)
+
+
+def _inplace_replace_geometry_filter_name(
+        node: pygeofilter.ast.Node,
+        geometry_column_name: str
+):
+    """Recursively traverse node tree and rename nodes of type ``Attribute``.
+
+    Nodes of type ``Attribute`` named ``geometry`` are renamed to the value of
+    the ``geometry_column_name`` parameter.
+    """
+    try:
+        sub_nodes = node.get_sub_nodes()
+    except AttributeError:
+        pass
+    else:
+        for sub_node in sub_nodes:
+            is_attribute_node = isinstance(sub_node, pygeofilter.ast.Attribute)
+            if is_attribute_node and sub_node.name == "geometry":
+                sub_node.name = geometry_column_name
+            else:
+                _inplace_replace_geometry_filter_name(
+                    sub_node, geometry_column_name)
